# -*- Encoding: utf-8 -*-
'''wrappers.py - Wrappers for Java classes and instances

python-javabridge is licensed under the BSD license.  See the
accompanying file LICENSE for details.

Copyright (c) 2003-2009 Massachusetts Institute of Technology
Copyright (c) 2009-2014 Broad Institute
All rights reserved.

'''

import inspect
import sys
import numpy as np
import javabridge as J

try:
    basestring            # Python 2
except NameError:
    basestring = (str, )  # Python 3


class JWrapper(object):
    '''A class that wraps a Java object
    
    JWrapper uses Java reflection to find a Java object's methods and fields.
    You can then use dot notation to call the methods and get references
    to the fields. If methods return Java objects, these will also be
    wrapped, so you can use the wrapper to do almost anything.
    
    When a class has overloaded methods with the same name, JWrapper will
    try to pick the one that matches the types of the arguments.
    
    To access static methods and fields, use JClassWrapper.
    
    `self.o` is the JB_Object wrapped by the wrapper. You
    can use `self.o` as an argument for the collection wrappers or anywhere
    else you might use a JB_Object.
    
    Usage:
    
        >>> a = JWrapper(javabridge.make_instance("java/util/ArrayList", "()V"))
        >>> a.add("Hello")
        >>> a.add("World")
        >>> a.size()
        2
        >>> a.get(0).toLowerCase()
        hello
    '''
    def __init__(self, o):
        '''Initialize the JWrapper with a Java object
        
        :param o: a Java object (class = JB_Object)
        '''
        STATIC = J.get_static_field("java/lang/reflect/Modifier", "STATIC", "I")
        self.o = o
        self.class_wrapper = J.get_class_wrapper(o)
        env = J.get_env()
        jmethods = env.get_object_array_elements(self.class_wrapper.getMethods())
        methods = {}
        for jmethod in jmethods:
            if (J.call(jmethod, "getModifiers", "()I") & STATIC) == STATIC:
                continue
            method = J.get_method_wrapper(jmethod)
            name = method.getName()
            if name not in methods:
                methods[name] = []
                fn = lambda naame=name: lambda *args: self.__call(naame, *args)
                fn = fn()
                fn.__doc__ = J.to_string(jmethod)
                setattr(self, name, fn)
            else:
                fn = getattr(self, name)
                fn.__doc__ = fn.__doc__ +"\n"+J.to_string(jmethod)
            methods[name].append(method)
        jfields = env.get_object_array_elements(
            self.class_wrapper.getFields(self))
        field_class = env.find_class("java/lang/reflect/Field")
        method_id = env.get_method_id(
            field_class, "getName", "()Ljava/lang/String;")
        self.field_names = [ 
            env.get_string_utf(env.call_method(o, method_id)) for o in jfields]
        self.methods = methods
        
    def __getattr__(self, name):
        if name in ("o", "class_wrapper", "methods", "field_names"):
            raise AttributeError()
        if not hasattr(self, "methods") or not hasattr(self, "field_names"):
            # not initialized
            raise AttributeError()
        if name not in self.field_names:
            raise AttributeError()
        try:
            jfield = self.class_wrapper.getField(name)
        except:
            raise AttributeError()
    
        STATIC = J.get_static_field("java/lang/reflect/Modifier", "STATIC", "I")
        if (J.call(jfield, "getModifiers", "()I") & STATIC) == STATIC:
            raise AttributeError()
        klass = J.call(jfield, "getType", "()Ljava/lang/Class;")
        result = J.get_field(self.o, name, sig(klass))
        if isinstance(result, J.JB_Object):
            result = JWrapper(result)
        return result
    
    def __setattr__(self, name, value):
        if name in ("o", "class_wrapper", "methods", "field_names") or \
           not hasattr(self, "methods"):
            object.__setattr__(self, name, value)
            return
        try:
            jfield = self.class_wrapper.getField(name)
        except:
            object.__setattr__(self, name, value)
            return
    
        STATIC = J.get_static_field("java/lang/reflect/Modifier", "STATIC", "I")
        if (J.call(jfield, "getModifiers", "()I") & STATIC) == STATIC:
            raise AttributeError()
        klass = J.call(jfield, "getType", "()Ljava/lang/Class;")
        result = J.set_field(self.o, name, sig(klass), value)
            
    def __call(self, method_name, *args):
        '''Call the appropriate overloaded method with the given name
        
        :param method_name: the name of the method to call
        :param *args: the arguments to the method, which are used to
                      disambiguate between similarly named methods
        '''
        env = J.get_env()
        last_e = None
        for method in self.methods[method_name]:
            params = env.get_object_array_elements(method.getParameterTypes())
            is_var_args = J.call(method.o, "isVarArgs", "()Z")
            if len(args) < len(params) - (1 if is_var_args else 0):
                continue
            if len(args) > len(params) and not is_var_args:
                continue
            if is_var_args:
                pm1 = len(params)-1
                args1 = list(args[:pm1]) + [args[pm1:]]
            else:
                args1 = args
            try:
                cargs = [cast(o, klass) for o, klass in zip(args1, params)]
            except:
                last_e = sys.exc_info()[1]
                continue
            rtype = J.call(method.o, "getReturnType", "()Ljava/lang/Class;")
            args_sig = "".join(map(sig, params))
            rsig = sig(rtype)
            msig = "(%s)%s" % (args_sig, rsig)
            result =  J.call(self.o, method_name, msig, *cargs)
            if isinstance(result, J.JB_Object):
                result = JWrapper(result)
            return result
        raise TypeError("No matching method found for %s" % method_name)
    
    def __repr__(self):
        classname = J.call(J.call(self.o, "getClass", "()Ljava/lang/Class;"), 
                           "getName", "()Ljava/lang/String;")
        return "Instance of %s: %s" % (classname, J.to_string(self.o))
    
    def __str__(self):
        return J.to_string(self.o)
    
    def __int__(self):
        return self.intValue()

    def __float__(self):
        return self.floatValue()

    def __len__(self):
        if not J.is_instance_of(self.o,'java/util/Collection'):
            raise TypeError("%s is not a Collection and does not support __len__" % self)
        return self.size()

    def __getitem__(self, i):
        if not J.is_instance_of(self.o,'java/util/Collection'):
            raise TypeError("%s is not a Collection and does not support __getitem__" % self)
        return self.get(i)

    def __setitem__(self, i, v):
        if not J.is_instance_of(self.o,'java/util/Collection'):
            raise TypeError("%s is not a Collection and does not support __setitem__" % self)
        return self.set(i, v) 
    
    class Iterator:
        def __init__(self, o):
            self.o = o
            self.i = 0

        def next(self):
            return self.__next__()

        def __next__(self):
            if self.i == len(self.o):
                raise StopIteration
            self.i = self.i +1
            return self.o[self.i-1]

    def __iter__(self):
        if not J.is_instance_of(self.o,'java/util/Collection'):
            raise TypeError("%s is not a Collection and does not support __iter__" % self)
        return self.Iterator(self)

class JClassWrapper(object):
    '''Wrapper for a class
    
    JWrapper uses Java reflection to find a Java object's methods and fields.
    You can then use dot notation to call the static methods and get references
    to the static fields. If methods return Java objects, these will also be
    wrapped, so you can use the wrapper to do almost anything.
    
    When a class has overloaded methods with the same name, JWrapper will
    try to pick the one that matches the types of the arguments.
    
    >>> Integer = JClassWrapper("java.lang.Integer")
    >>> Integer.MAX_VALUE
    2147483647
    '''
    def __init__(self, class_name):
        '''Initialize to wrap a class name
        
        :param class_name: name of class in dotted form, e.g. java.lang.Integer
        '''
        STATIC = J.get_static_field("java/lang/reflect/Modifier", "STATIC", "I")
        self.cname = class_name.replace(".", "/")
        self.klass = J.get_class_wrapper(J.class_for_name(class_name), True)
        self.static_methods = {}
        env = J.get_env()
        jmethods = env.get_object_array_elements(self.klass.getMethods())
        methods = {}
        for jmethod in jmethods:
            if (J.call(jmethod, "getModifiers", "()I") & STATIC) != STATIC:
                continue
            method = J.get_method_wrapper(jmethod)
            name = method.getName()
            if name not in methods:
                methods[name] = []
                fn = lambda naame=name: lambda *args: self.__call_static(naame, *args)
                fn = fn()
                fn.__doc__ = J.to_string(jmethod)
                setattr(self, name, fn)
            else:
                fn = getattr(self, name)
                fn.__doc__ = fn.__doc__ +"\n"+J.to_string(jmethod)
            methods[name].append(method)
        jfields = env.get_object_array_elements(self.klass.getFields(self))
        field_class = env.find_class("java/lang/reflect/Field")
        method_id = env.get_method_id(
            field_class, "getName", "()Ljava/lang/String;")
        self.field_names = [ 
            env.get_string_utf(env.call_method(o, method_id)) for o in jfields]
        self.methods = methods
        
    def __getattr__(self, name):
        if name in ("klass", "static_methods", "methods", "cname", 
                    "field_names"):
            raise AttributeError()
        if not hasattr(self, "methods") or not hasattr(self, "field_names"):
            raise AttributeError()
        if name not in self.field_names:
            raise AttributeError("Cound not find field %s" % name)
        try:
            jfield = self.klass.getField(name)
        except:
            raise AttributeError("Could not find field %s" % name)
    
        STATIC = J.get_static_field("java/lang/reflect/Modifier", "STATIC", "I")
        if (J.call(jfield, "getModifiers", "()I") & STATIC) != STATIC:
            raise AttributeError("Field %s is not static" % name)
        klass = J.call(jfield, "getType", "()Ljava/lang/Class;")
        result = J.get_static_field(self.cname, name, sig(klass))
        if isinstance(result, J.JB_Object):
            result = JWrapper(result)
        return result
    
    def __setattr__(self, name, value):
        if name in ("klass", "static_methods", "methods", "cname", 
                    "field_names") or not hasattr(self, "methods"):
            object.__setattr__(self, name, value)
            return
        try:
            jfield = self.klass.getField(name)
        except:
            return object.__setattr__(self, name, value)

        STATIC = J.get_static_field("java/lang/reflect/Modifier", "STATIC", "I")
        if (J.call(jfield, "getModifiers", "()I") & STATIC) != STATIC:
            raise AttributeError()
        klass = J.call(jfield, "getType", "()Ljava/lang/Class;")
        result = J.set_static_field(self.cname, name, sig(klass), value)
    
    def __call_static(self, method_name, *args):
        '''Call the appropriate overloaded method with the given name
        
        :param method_name: the name of the method to call
        :param *args: the arguments to the method, which are used to
                      disambiguate between similarly named methods
        '''
        env = J.get_env()
        last_e = None
        for method in self.methods[method_name]:
            params = env.get_object_array_elements(method.getParameterTypes())
            is_var_args = J.call(method.o, "isVarArgs", "()Z")
            if len(args) < len(params) - (1 if is_var_args else 0):
                continue
            if len(args) > len(params) and not is_var_args:
                continue
            if is_var_args:
                pm1 = len(params)-1
                args1 = list(args[:pm1]) + [args[pm1:]]
            else:
                args1 = args
            try:
                cargs = [cast(o, klass) for o, klass in zip(args1, params)]
            except:
                last_e = sys.exc_info()[1]
                continue
            rtype = J.call(method.o, "getReturnType", "()Ljava/lang/Class;")
            args_sig = "".join(map(sig, params))
            rsig = sig(rtype)
            msig = "(%s)%s" % (args_sig, rsig)
            result =  J.static_call(self.cname, method_name, msig, *cargs)
            if isinstance(result, J.JB_Object):
                result = JWrapper(result)
            return result
        raise TypeError("No matching method found for %s" % method_name)

    def __call__(self, *args):
        '''Constructors'''
        env = J.get_env()
        jconstructors = self.klass.getConstructors()
        for jconstructor in env.get_object_array_elements(jconstructors):
            constructor = J.get_constructor_wrapper(jconstructor)
            params = env.get_object_array_elements(
                constructor.getParameterTypes())
            is_var_args = J.call(constructor.o, "isVarArgs", "()Z")
            if len(args) < len(params) - (1 if is_var_args else 0):
                continue
            if len(args) > len(params) and not is_var_args:
                continue
            if is_var_args:
                pm1 = len(params)-1
                args1 = list(args[:pm1]) + [args[pm1:]]
            else:
                args1 = args
            try:
                cargs = [cast(o, klass) for o, klass in zip(args1, params)]
            except:
                last_e = sys.exc_info()[1]
                continue
            args_sig = "".join(map(sig, params))
            msig = "(%s)V" % (args_sig)
            result =  J.make_instance(self.cname, msig, *cargs)
            result = JWrapper(result)
            return result
        raise TypeError("No matching constructor found")
    
class JProxy(object):
    '''A wrapper around java.lang.reflect.Proxy
    
    The wrapper takes a dictionary of either method name or a
    `java.lang.reflect.Method` instance to a callable that handles
    the method. You can also subclass JProxy and define methods
    with the same names as the Java methods and they will be called.    

    An example:

        >>> import javabridge
        >>> import sys
        >>> runnable = javabridge.JProxy(
                'java.lang.Runnable',
                dict(run=lambda:sys.stderr.write("Hello, world.\\n"))))
        >>> javabridge.JWrapper(runnable.o).run()
        
    Another example:

        >>> import javabridge
        >>> import sys
        >>> class MyRunnable(javabridge.JProxy):
                def __init__(self):
                    javabridge.JProxy.__init__(self, 'java.lang.Runnable')
                def run(self):
                    sys.stderr.write("Hello, world.\\n")
        >>> proxy = MyRunnable()
        >>> javabridge.JWrapper(runnable.o).run()
    '''
    def __init__(self, base_class_name, d=None):
        '''Initialize the proxy with the interface name and methods
        
        :param base_class_name: the class name of the interface to implement
                                in dotted form (e.g. java.lang.Runnable)
        :param d: an optional dictionary of method name to implementation
        '''
        self.ref_id, self.ref = J.create_jref(self)
        self.__d = d or {}
        jclass = J.class_for_name(base_class_name)
        loader = J.call(jclass, "getClassLoader",
                        "()Ljava/lang/ClassLoader;")
        env = J.get_env()
        classes = env.make_object_array(1, env.find_class("java/lang/Class"))
        env.set_object_array_element(classes, 0, jclass)
        handler = J.make_instance(
            "org/cellprofiler/javabridge/CPythonInvocationHandler",
            "(Ljava/lang/String;)V", self.ref_id)
        self.o = J.static_call(
            "java/lang/reflect/Proxy",
            "newProxyInstance",
            "(Ljava/lang/ClassLoader;"
            "[Ljava/lang/Class;"
            "Ljava/lang/reflect/InvocationHandler;)"
            "Ljava/lang/Object;",
            loader, classes, handler)
        
    def __call__(self, proxy, method, jargs):
        name = J.call(method, "getName", "()Ljava/lang/String;")
        env = J.get_env()
        args = tuple(env.get_object_array_elements(jargs))
        if name in self.__d:
            result = self.__d[name](*args)
        else:
            result = getattr(self, name)(*args)
        retclass = J.call(method, "getReturnType", "()Ljava/lang/Class;")
        return cast(result, retclass)
    
def importClass(class_name, import_name = None):
    '''Import a wrapped class into the global context
    
    :param class_name: a dotted class name such as java.lang.String
    :param import_name: if defined, use this name instead of the class's name
    '''
    if import_name is None:
        if "." in class_name:
            import_name = class_name.rsplit(".", 1)[1]
        else:
            import_name = class_name
    frame = inspect.currentframe(1)
    frame.f_locals[import_name] = JClassWrapper(class_name)

def sig(klass):
    '''Return the JNI signature for a class'''
    name = J.call(klass, "getName", "()Ljava/lang/String;")
    if not (J.call(klass, "isPrimitive", "()Z") or 
            J.call(klass, "isArray", "()Z")):
        name = "L%s;" % name
    if name == 'void':
        return "V"
    if name == 'int':
        return "I"
    if name == 'byte':
        return "B"
    if name == 'boolean':
        return "Z"
    if name == 'long':
        return "J"
    if name == 'float':
        return "F"
    if name == 'double':
        return "D"
    if name == 'char':
        return "C"
    if name == 'short':
        return "S"
    return name.replace(".", "/")

def cast(o, klass):
    '''Cast the given object to the given class
    
    :param o: either a Python object or Java object to be cast
    :param klass: a java.lang.Class indicating the target class
    
    raises a TypeError if the object can't be cast.
    '''
    if J.call(klass, "getName", "()Ljava/lang/String;") == 'void':
        return None
    is_primitive = J.call(klass, "isPrimitive", "()Z")
    csig = sig(klass)
    if o is None:
        if not is_primitive:
            return None
        else:
            raise TypeError("Can't cast None to a primitive type")
        
    if isinstance(o, J.JB_Object):
        if J.call(klass, "isInstance", "(Ljava/lang/Object;)Z", o):
            return o
        classname = J.run_script("o.getClass().getCanonicalName()", dict(o=o))
        klassname = J.run_script("klass.getCanonicalName()", dict(klass=klass))
        raise TypeError("Object of class %s cannot be cast to %s",
                        classname, klassname)
    elif hasattr(o, "o"):
        return cast(o.o, klass)
    elif not np.isscalar(o):
        component_type = J.call(klass, "getComponentType", "()Ljava/lang/Class;")
        if component_type is None:
            raise TypeError("Argument must not be a sequence")
        if len(o) > 0:
            # Test if an element can be cast to the array type
            cast(o[0], component_type)
        return J.get_nice_arg(o, csig)
    elif is_primitive or csig in \
         ('Ljava/lang/String;', 'Ljava/lang/CharSequence;', 
          'Ljava/lang/Object;'):
        if csig == 'Ljava/lang/CharSequence;':
            csig = 'Ljava/lang/String;'
        elif csig == 'C' and isinstance(o, basestring) and len(o) != 1:
            raise TypeError("Failed to convert string of length %d to char" %
                            len(o))
        return J.get_nice_arg(o, csig)
    raise TypeError("Failed to convert argument to %s" % csig)

all = [JWrapper, JClassWrapper]



"""locate.py - determine architecture and find Java

python-javabridge is licensed under the BSD license.  See the
accompanying file LICENSE for details.

Copyright (c) 2003-2009 Massachusetts Institute of Technology
Copyright (c) 2009-2013 Broad Institute
All rights reserved.lo

"""

import ctypes
import os
import sys
import logging
import subprocess
import re

# Note: if a matching gcc is available from the shell on Windows, its
#       probably safe to assume the user is in an MINGW or MSYS or Cygwin
#       environment, in which case he/she wants to compile with gcc for
#       Windows, in which case the correct compiler flags will be triggered
#       by is_mingw. This method is not great, improve it if you know a
#       better way to discriminate between compilers on Windows.
def is_mingw():
    # currently this check detects mingw only on Windows. Extend for other
    # platforms if required:
    if (os.name != "nt"):
        return False

    # if the user defines DISTUTILS_USE_SDK or MSSdk, we expect they want
    # to use Microsoft's compiler (as described here:
    # https://github.com/cython/cython/wiki/CythonExtensionsOnWindows):
    if (os.getenv("DISTUTILS_USE_SDK") != None or os.getenv("MSSdk") != None):
        return False

    mingw32 = ""
    mingw64 = ""
    if (os.getenv("MINGW32_PREFIX")):
        mingw32 = os.getenv("MINGW32_PREFIX")
    if (os.getenv("MINGW64_PREFIX")):
        mingw64 = os.getenv("MINGW64_PREFIX")

    # if any invocation of gcc works, then we assume the user wants mingw:
    test = "gcc --version > NUL 2>&1"
    if (os.system(test) == 0 or os.system(mingw32+test) == 0 or os.system(mingw64+test) == 0):
        return True

    return False



is_linux = sys.platform.startswith('linux')
is_mac = sys.platform == 'darwin'
is_win = sys.platform.startswith("win")
is_win64 = (is_win and (os.environ["PROCESSOR_ARCHITECTURE"] == "AMD64"))
is_msvc = (is_win and
           ((sys.version_info.major == 2 and sys.version_info.minor >= 6) or
            (sys.version_info.major == 3)))
is_mingw = is_mingw()

if is_win:
    if sys.version_info.major == 2:
        import _winreg as winreg
        from exceptions import WindowsError
    else:
        import winreg

logger = logging.getLogger(__name__)

def find_javahome():
    """Find JAVA_HOME if it doesn't exist"""
    if 'JAVA_HOME' in os.environ:
        return os.environ['JAVA_HOME']
    elif is_mac:
        # Use the "java_home" executable to find the location
        # see "man java_home"
        libc = ctypes.CDLL("/usr/lib/libc.dylib")
        if sys.maxsize <= 2**32:
            arch = "i386"
        else:
            arch = "x86_64"
        try:
            result = subprocess.check_output(["/usr/libexec/java_home", "--arch", arch])
            path = result.strip().decode("utf-8")
            for place_to_look in (
                os.path.join(os.path.dirname(path), "Libraries"),
                os.path.join(path, "jre", "lib", "server")):
                # In "Java for OS X 2015-001" libjvm.dylib is a symlink to libclient.dylib
                # which is i686 only, whereas libserver.dylib contains both architectures.
                for file_to_look in ('libjvm.dylib',
                                     'libclient.dylib',
                                     'libserver.dylib'):
                    lib = os.path.join(place_to_look, file_to_look)
                    #
                    # dlopen_preflight checks to make sure the dylib
                    # can be loaded in the current architecture
                    #
                    if os.path.exists(lib) and \
                       libc.dlopen_preflight(lib.encode('utf-8')) != 0:
                        return path
            else:
                logger.error("Could not find Java JRE compatible with %s architecture" % arch)
                if arch == "i386":
                    logger.error(
                        "Please visit https://support.apple.com/kb/DL1572 for help\n"
                        "installing Apple legacy Java 1.6 for 32 bit support.")
                return None
        except:
            logger.error("Failed to run /usr/libexec/java_home, defaulting to best guess for Java", exc_info=1)
        return "/System/Library/Frameworks/JavaVM.framework/Home"
    elif is_linux:
        def get_out(cmd):
            p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            o, ignore = p.communicate()
            if p.poll() != 0:
                raise Exception("Error finding javahome on linux: %s" % cmd)
            o = o.strip().decode('utf-8')
            return o
        java_bin = get_out(["bash", "-c", "type -p java"])
        java_dir = get_out(["readlink", "-f", java_bin])
        java_version_string = get_out(["bash", "-c", "java -version"])
        if re.search('^openjdk', java_version_string, re.MULTILINE) is not None:
            jdk_dir = os.path.join(java_dir, "..", "..", "..")
        elif re.search('^java', java_version_string, re.MULTILINE) is not None:
            jdk_dir = os.path.join(java_dir, "..", "..")
        else:
            raise RuntimeError(
                "Failed to determine JDK vendor. "
                "OpenJDK and Oracle JDK are supported."
            )
        jdk_dir = os.path.abspath(jdk_dir)
        return jdk_dir
    elif is_win:
        # Registry keys changed in 1.9
        # https://docs.oracle.com/javase/9/migrate/toc.htm#GUID-EEED398E-AE37-4D12-AB10-49F82F720027
        java_key_paths = (
            'SOFTWARE\\JavaSoft\\JRE',
            'SOFTWARE\\JavaSoft\\Java Runtime Environment',
            'SOFTWARE\\JavaSoft\\JDK'
        )
        for java_key_path in java_key_paths:
            looking_for = java_key_path
            try:
                kjava = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, java_key_path)
                looking_for = java_key_path + "\\CurrentVersion"
                kjava_values = dict([winreg.EnumValue(kjava, i)[:2]
                                     for i in range(winreg.QueryInfoKey(kjava)[1])])
                current_version = kjava_values['CurrentVersion']
                looking_for = java_key_path + '\\' + current_version
                kjava_current = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,
                                                looking_for)
                kjava_current_values = dict([winreg.EnumValue(kjava_current, i)[:2]
                                             for i in range(winreg.QueryInfoKey(kjava_current)[1])])
                return kjava_current_values['JavaHome']
            except WindowsError as e:
                if e.errno == 2:
                    continue
                else:
                    raise

        raise RuntimeError(
            "Failed to find the Java Runtime Environment. "
            "Please download and install the Oracle JRE 1.6 or later"
        )


def find_jdk():
    """Find the JDK under Windows"""
    if 'JDK_HOME' in os.environ:
        return os.environ['JDK_HOME']
    if is_linux:
        jdk_home = find_javahome()
        if jdk_home.endswith("jre") or jdk_home.endswith("jre/"):
            jdk_home = jdk_home[:jdk_home.rfind("jre")]
        return jdk_home
    if is_mac:
        return find_javahome()
    if is_win:
        # Registry keys changed in 1.9
        # https://docs.oracle.com/javase/9/migrate/toc.htm#GUID-EEED398E-AE37-4D12-AB10-49F82F720027
        jdk_key_paths = (
            'SOFTWARE\\JavaSoft\\JDK',
            'SOFTWARE\\JavaSoft\\Java Development Kit',
        )
        for jdk_key_path in jdk_key_paths:
            try:
                kjdk = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, jdk_key_path)
                kjdk_values = dict([winreg.EnumValue(kjdk, i)[:2]
                                     for i in range(winreg.QueryInfoKey(kjdk)[1])])
                current_version = kjdk_values['CurrentVersion']
                kjdk_current = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE,
                                               jdk_key_path + '\\' + current_version)
                kjdk_current_values = dict([winreg.EnumValue(kjdk_current, i)[:2]
                                            for i in range(winreg.QueryInfoKey(kjdk_current)[1])])
                return kjdk_current_values['JavaHome']
            except WindowsError as e:
                if e.errno == 2:
                    continue
                else:
                    raise

        raise RuntimeError(
            "Failed to find the Java Development Kit. "
            "Please download and install the Oracle JDK 1.6 or later"
        )

def find_javac_cmd():
    """Find the javac executable"""
    if is_win:
        jdk_base = find_jdk()
        javac = os.path.join(jdk_base, "bin", "javac.exe")
        if os.path.isfile(javac):
            return javac
        raise RuntimeError("Failed to find javac.exe in its usual location under the JDK (%s)" % javac)
    else:
        # will be along path for other platforms
        return "javac"

def find_jar_cmd():
    """Find the javac executable"""
    if is_win:
        jdk_base = find_jdk()
        javac = os.path.join(jdk_base, "bin", "jar.exe")
        if os.path.isfile(javac):
            return javac
        raise RuntimeError("Failed to find jar.exe in its usual location under the JDK (%s)" % javac)
    else:
        # will be along path for other platforms
        return "jar"


def find_jre_bin_jdk_so():
    """Finds the jre bin dir and the jdk shared library file"""
    jvm_dir = None
    java_home = find_javahome()
    if java_home is not None:
        found_jvm = False
        for jre_home in (java_home, os.path.join(java_home, "jre"), os.path.join(java_home, 'default-java')):
            jre_bin = os.path.join(jre_home, 'bin')
            jre_libexec = os.path.join(jre_home, 'bin' if is_win else 'lib')
            arches = ('amd64', 'i386', '') if is_linux else ('',)
            lib_prefix = '' if is_win else 'lib'
            lib_suffix = '.dll' if is_win else ('.dylib' if is_mac else '.so')
            for arch in arches:
                for place_to_look in ('client','server'):
                    jvm_dir = os.path.join(jre_libexec, arch, place_to_look)
                    jvm_so = os.path.join(jvm_dir, lib_prefix + "jvm" + lib_suffix)
                    if os.path.isfile(jvm_so):
                        return (jre_bin, jvm_so)
    return (jre_bin, None)



'''test_javabridge.py - test the low-level interface

python-javabridge is licensed under the BSD license.  See the
accompanying file LICENSE for details.

Copyright (c) 2003-2009 Massachusetts Institute of Technology
Copyright (c) 2009-2013 Broad Institute
All rights reserved.

'''

__version__="$Revision$"

import os
import numpy as np
import unittest

import javabridge
jb = javabridge


class TestJavabridge(unittest.TestCase):

    def setUp(self):
        self.env = javabridge.attach()

    def tearDown(self):
        javabridge.detach()

    def test_01_01_version(self):
        major,minor = self.env.get_version()
        
    def test_01_02_find_class(self):
        string_class = self.env.find_class('java/lang/String')
        self.assertTrue(isinstance(string_class, jb.JB_Class))

    def test_01_03_00_new_string_utf(self):
        jstring = self.env.new_string_utf("Hello, world")
        self.assertTrue(isinstance(jstring, jb.JB_Object))
        
    def test_01_03_01_new_string_unicode(self):
        s = u"Hola ni\u00F1os"
        jstring = self.env.new_string(s)
        self.assertEqual(self.env.get_string_utf(jstring), s)
        
    def test_01_03_02_new_string_string(self):
        s = "Hello, world"
        jstring = self.env.new_string(s)
        self.assertEqual(self.env.get_string_utf(jstring), s)
        
    def test_01_03_03_new_string_zero_length(self):
        jstring = self.env.new_string(u"")
        self.assertEqual(self.env.get_string_utf(jstring), "")
        
    def test_01_04_00_get_string_utf(self):
        jstring = self.env.new_string_utf("Hello, world")
        pstring = self.env.get_string_utf(jstring)
        self.assertEqual(pstring, "Hello, world")

    def test_01_04_01_get_string(self):
        s = u"Hola ni\u00F1os"
        jstring = self.env.new_string(s)
        self.assertTrue(self.env.get_string(jstring), s)        
        
    def test_01_05_get_object_class(self):
        jstring = self.env.new_string_utf("Hello, world")
        string_class = self.env.get_object_class(jstring)
        self.assertTrue(isinstance(string_class, jb.JB_Class))
        
    def test_01_06_deallocate_object(self):
        jstring = self.env.new_string_utf("Hello, world")
        del jstring
        
    def test_01_09_get_method_id(self):
        klass = self.env.find_class("java/lang/String")
        method_id = self.env.get_method_id(klass,'charAt','(I)C')
        self.assertTrue(method_id is not None)
        
    def test_01_10_get_static_method_id(self):
        klass = self.env.find_class("java/lang/String")
        method_id = self.env.get_static_method_id(klass, 'copyValueOf','([C)Ljava/lang/String;')
        self.assertTrue(method_id is not None)
        
    def test_01_11_new_object(self):
        klass = self.env.find_class("java/lang/Byte")
        method_id = self.env.get_method_id(klass, '<init>','(Ljava/lang/String;)V')
        jbyte = self.env.new_object(klass, method_id, self.env.new_string_utf("55"))
        self.assertTrue(jbyte is not None)
        
    def test_01_11_01_is_instance_of(self):
        klassByte = self.env.find_class("java/lang/Byte")
        method_id = self.env.get_method_id(klassByte, '<init>','(Ljava/lang/String;)V')
        jbyte = self.env.new_object(klassByte, method_id, self.env.new_string_utf("55"))
        klassNumber = self.env.find_class("java/lang/Number")
        self.assertTrue(self.env.is_instance_of(jbyte, klassNumber))
        
    def test_01_11_02_isnt_instance_of(self):
        klassByte = self.env.find_class("java/lang/Byte")
        method_id = self.env.get_method_id(klassByte, '<init>','(Ljava/lang/String;)V')
        jbyte = self.env.new_object(klassByte, method_id, self.env.new_string_utf("55"))
        klassString = self.env.find_class("java/lang/String")
        self.assertFalse(self.env.is_instance_of(jbyte, klassString))
    
    def test_01_12_get_static_field_id(self):
        klass = self.env.find_class("java/lang/Boolean")
        field_id = self.env.get_static_field_id(klass, "FALSE","Ljava/lang/Boolean;")
        self.assertTrue(field_id is not None)
        
    def test_01_13_get_byte_array_elements(self):
        pass # see test_03_09_call_method_array for test
    
    def test_01_14_get_object_array_elements(self):
        jstring = self.env.new_string_utf("Hello, world")
        klass = self.env.get_object_class(jstring)
        method_id = self.env.get_method_id(klass, 'split', '(Ljava/lang/String;)[Ljava/lang/String;')
        split = self.env.new_string_utf(", ")
        result = self.env.call_method(jstring, method_id, split)
        result = self.env.get_object_array_elements(result)
        self.assertEqual(len(result), 2)
        self.assertEqual(self.env.get_string_utf(result[0]), "Hello")
        self.assertEqual(self.env.get_string_utf(result[1]), "world")
        
    def test_01_15_make_byte_array(self):
        array = np.array([ord(x) for x in "Hello, world"],np.uint8)
        jarray = self.env.make_byte_array(array)
        klass = self.env.find_class("java/lang/String")
        method_id = self.env.get_method_id(klass, '<init>', '([B)V')
        result = self.env.new_object(klass, method_id, jarray)
        self.assertEqual(self.env.get_string_utf(result), "Hello, world")
        
    def test_01_16_get_array_length(self):
        jstring = self.env.new_string_utf("Hello, world")
        klass = self.env.get_object_class(jstring)
        method_id = self.env.get_method_id(klass, 'split', '(Ljava/lang/String;)[Ljava/lang/String;')
        split = self.env.new_string_utf(", ")
        result = self.env.call_method(jstring, method_id, split)
        self.assertEqual(self.env.get_array_length(result), 2)
        
    def test_01_17_make_object_array(self):
        klass = self.env.find_class("java/lang/String")
        jarray = self.env.make_object_array(15, klass)
        length = self.env.get_array_length(jarray)
        self.assertEqual(length, 15)
        
    def test_01_18_set_object_array_element(self):
        klass = self.env.find_class("java/lang/String")
        jarray = self.env.make_object_array(15, klass)
        for i in range(15):
            v = self.env.new_string_utf(str(i))
            self.env.set_object_array_element(jarray, i, v)
        result = self.env.get_object_array_elements(jarray)
        self.assertEqual(len(result), 15)
        for i, elem in enumerate(result):
            v = self.env.get_string_utf(elem)
            self.assertEqual(str(i), v)
            
    def test_01_19_0_make_boolean_array(self):
        np.random.seed(1190)
        array = np.random.uniform(size=105) > .5
        jarray = self.env.make_boolean_array(array)
        result = self.env.get_boolean_array_elements(jarray)
        self.assertTrue(np.all(array == result))
            
    def test_01_19_make_short_array(self):
        np.random.seed(119)
        array = (np.random.uniform(size=10) * 65535 - 32768).astype(np.int16)
        array = np.unique(array)
        array.sort()
        jarray = self.env.make_short_array(array)
        klass = self.env.find_class("java/util/Arrays")
        method_id = self.env.get_static_method_id(klass, "binarySearch",
                                                  "([SS)I")
        for i, value in enumerate(array):
            self.assertEqual(i, self.env.call_static_method(
                klass, method_id, jarray, array[i]))
            
    def test_01_20_make_int_array(self):
        np.random.seed(120)
        array = (np.random.uniform(size=10) * (2.0 ** 32-1) - (2.0 ** 31)).astype(np.int32)
        array = np.unique(array)
        array.sort()
        jarray = self.env.make_int_array(array)
        klass = self.env.find_class("java/util/Arrays")
        method_id = self.env.get_static_method_id(klass, "binarySearch",
                                                  "([II)I")
        for i, value in enumerate(array):
            self.assertEqual(i, self.env.call_static_method(
                klass, method_id, jarray, array[i]))
            
    def test_01_21_make_long_array(self):
        np.random.seed(121)
        array = (np.random.uniform(size=10) * (2.0 ** 64) - (2.0 ** 63)).astype(np.int64)
        array = np.unique(array)
        array.sort()
        jarray = self.env.make_long_array(array)
        klass = self.env.find_class("java/util/Arrays")
        method_id = self.env.get_static_method_id(klass, "binarySearch",
                                                  "([JJ)I")
        for i, value in enumerate(array):
            self.assertEqual(i, self.env.call_static_method(
                klass, method_id, jarray, array[i]))
            
    def test_01_22_make_float_array(self):
        np.random.seed(122)
        array = np.random.uniform(size=10).astype(np.float32)
        array = np.unique(array)
        array.sort()
        jarray = self.env.make_float_array(array)
        klass = self.env.find_class("java/util/Arrays")
        method_id = self.env.get_static_method_id(klass, "binarySearch",
                                                  "([FF)I")
        for i, value in enumerate(array):
            self.assertEqual(i, self.env.call_static_method(
                klass, method_id, jarray, array[i]))
            
    def test_01_23_make_double_array(self):
        np.random.seed(123)
        array = np.random.uniform(size=10).astype(np.float64)
        array = np.unique(array)
        array.sort()
        jarray = self.env.make_double_array(array)
        klass = self.env.find_class("java/util/Arrays")
        method_id = self.env.get_static_method_id(klass, "binarySearch",
                                                  "([DD)I")
        for i, value in enumerate(array):
            self.assertEqual(i, self.env.call_static_method(
                klass, method_id, jarray, array[i]))
            
    def test_01_24_get_short_array_elements(self):
        np.random.seed(124)
        array = (np.random.uniform(size=10) * 65535 - 32768).astype(np.int16)
        jarray = self.env.make_short_array(array)
        result = self.env.get_short_array_elements(jarray)
        self.assertTrue(np.all(array == result))
            
    def test_01_25_get_int_array_elements(self):
        np.random.seed(125)
        array = (np.random.uniform(size=10) * (2.0 ** 32-1) - (2.0 ** 31)).astype(np.int32)
        jarray = self.env.make_int_array(array)
        result = self.env.get_int_array_elements(jarray)
        self.assertTrue(np.all(array == result))
            
    def test_01_26_get_long_array_elements(self):
        np.random.seed(126)
        array = (np.random.uniform(size=10) * (2.0 ** 64) - (2.0 ** 63)).astype(np.int64)
        jarray = self.env.make_long_array(array)
        result = self.env.get_long_array_elements(jarray)
        self.assertTrue(np.all(array == result))
            
    def test_01_27_get_float_array_elements(self):
        np.random.seed(127)
        array = np.random.uniform(size=10).astype(np.float32)
        jarray = self.env.make_float_array(array)
        result = self.env.get_float_array_elements(jarray)
        self.assertTrue(np.all(array == result))
            
    def test_01_28_get_double_array_elements(self):
        np.random.seed(128)
        array = np.random.uniform(size=10).astype(np.float64)
        jarray = self.env.make_double_array(array)
        result = self.env.get_double_array_elements(jarray)
        self.assertTrue(np.all(array == result))
            
    def test_02_01_exception_did_not_occur(self):
        self.assertTrue(self.env.exception_occurred() is None)
        
    def test_02_02_exception_occurred(self):
        klass = self.env.find_class("java/lang/String")
        self.env.get_method_id(klass,'copyValueOf','([C)Ljava/lang/String;')
        x = self.env.exception_occurred()
        self.assertTrue(isinstance(x, jb.JB_Object))
        self.env.exception_describe()
        self.env.exception_clear()
        self.assertTrue(self.env.exception_occurred() is None)
        
    def test_03_01_call_method_char(self):
        jstring = self.env.new_string_utf("Hello, world")
        klass = self.env.get_object_class(jstring)
        method_id = self.env.get_method_id(klass, 'charAt', '(I)C')
        
        for i, c in enumerate("Hello, world"):
            self.assertEqual(c, self.env.call_method(jstring, method_id, i))
    
    def test_03_02_call_method_bool(self):
        jstring = self.env.new_string_utf("Hello, world")
        klass = self.env.get_object_class(jstring)
        method_id = self.env.get_method_id(klass, 'equals', '(Ljava/lang/Object;)Z')
        self.assertTrue(self.env.call_method(jstring, method_id, jstring))
        self.assertFalse(self.env.call_method(jstring, method_id, self.env.new_string_utf("Foo")))
        
    def test_03_03_call_method_byte(self):
        klass = self.env.find_class("java/lang/Byte")
        method_id = self.env.get_method_id(klass, '<init>','(Ljava/lang/String;)V')
        jbyte = self.env.new_object(klass, method_id, self.env.new_string_utf("55"))
        method_id = self.env.get_method_id(klass, 'byteValue','()B')
        self.assertEqual(self.env.call_method(jbyte, method_id), 55)
    
    def test_03_04_call_method_short(self):
        klass = self.env.find_class("java/lang/Short")
        method_id = self.env.get_method_id(klass, '<init>','(Ljava/lang/String;)V')
        jshort = self.env.new_object(klass, method_id, self.env.new_string_utf("55"))
        method_id = self.env.get_method_id(klass, 'shortValue','()S')
        self.assertEqual(self.env.call_method(jshort, method_id), 55)
    
    def test_03_05_call_method_int(self):
        klass = self.env.find_class("java/lang/Integer")
        method_id = self.env.get_method_id(klass, '<init>','(Ljava/lang/String;)V')
        jint = self.env.new_object(klass, method_id, self.env.new_string_utf("65537"))
        method_id = self.env.get_method_id(klass, 'intValue','()I')
        self.assertEqual(self.env.call_method(jint, method_id), 65537)
    
    def test_03_06_call_method_long(self):
        klass = self.env.find_class("java/lang/Long")
        method_id = self.env.get_method_id(klass, '<init>','(Ljava/lang/String;)V')
        jlong = self.env.new_object(klass, method_id, self.env.new_string_utf("4611686018427387904"))
        method_id = self.env.get_method_id(klass, 'longValue','()J')
        self.assertEqual(self.env.call_method(jlong, method_id), 4611686018427387904)
    
    def test_03_07_call_method_float(self):
        klass = self.env.find_class("java/lang/Float")
        method_id = self.env.get_method_id(klass, '<init>','(Ljava/lang/String;)V')
        jfloat = self.env.new_object(klass, method_id, self.env.new_string_utf("55.3"))
        method_id = self.env.get_method_id(klass, 'floatValue','()F')
        self.assertAlmostEqual(self.env.call_method(jfloat, method_id), 55.3,5)
    
    def test_03_08_call_method_double(self):
        klass = self.env.find_class("java/lang/Double")
        method_id = self.env.get_method_id(klass, '<init>','(Ljava/lang/String;)V')
        jdouble = self.env.new_object(klass, method_id, self.env.new_string_utf("-55.64"))
        method_id = self.env.get_method_id(klass, 'doubleValue','()D')
        self.assertAlmostEqual(self.env.call_method(jdouble, method_id), -55.64)
        
    def test_03_09_call_method_array(self):
        s = "Hello, world"
        jstring = self.env.new_string_utf(s)
        klass = self.env.get_object_class(jstring)
        method_id = self.env.get_method_id(klass, 'getBytes', '()[B')
        result = self.env.call_method(jstring, method_id)
        self.assertTrue(isinstance(result, jb.JB_Object))
        a = self.env.get_byte_array_elements(result)
        self.assertEqual(np.array(s, "S%d" % len(s)).tostring(), a.tostring())
    
    def test_03_10_call_method_object(self):
        hello = self.env.new_string_utf("Hello, ")
        world = self.env.new_string_utf("world")
        klass = self.env.get_object_class(hello)
        method_id = self.env.get_method_id(klass, 'concat', '(Ljava/lang/String;)Ljava/lang/String;')
        result = self.env.call_method(hello, method_id, world)
        self.assertEqual("Hello, world", self.env.get_string_utf(result))
    
    def test_04_01_call_static_bool(self):
        klass = self.env.find_class("java/lang/Boolean")
        method_id = self.env.get_static_method_id(klass, "parseBoolean",'(Ljava/lang/String;)Z')
        self.assertTrue(method_id is not None)
        self.assertFalse(self.env.call_static_method(
            klass, method_id, self.env.new_string_utf("false")))
        self.assertTrue(self.env.call_static_method(
            klass, method_id, self.env.new_string_utf("true")))

    def test_04_02_call_static_byte(self):
        klass = self.env.find_class("java/lang/Byte")
        method_id = self.env.get_static_method_id(klass, "parseByte",'(Ljava/lang/String;)B')
        number = self.env.new_string_utf("55")
        self.assertEqual(self.env.call_static_method(klass, method_id, number), 55)
        
    def test_04_03_call_static_short(self):
        klass = self.env.find_class("java/lang/Short")
        method_id = self.env.get_static_method_id(klass, "parseShort",'(Ljava/lang/String;)S')
        number = self.env.new_string_utf("-55")
        self.assertEqual(self.env.call_static_method(klass, method_id, number), -55)
    
    def test_04_04_call_static_int(self):
        klass = self.env.find_class("java/lang/Integer")
        method_id = self.env.get_static_method_id(klass, "parseInt",'(Ljava/lang/String;)I')
        number = self.env.new_string_utf("55")
        self.assertEqual(self.env.call_static_method(klass, method_id, number), 55)
        
    def test_04_05_call_static_long(self):
        klass = self.env.find_class("java/lang/Long")
        method_id = self.env.get_static_method_id(klass, "parseLong",'(Ljava/lang/String;)J')
        number = self.env.new_string_utf("-55")
        self.assertEqual(self.env.call_static_method(klass, method_id, number), -55)
        
    def test_04_06_call_static_float(self):
        klass = self.env.find_class("java/lang/Float")
        method_id = self.env.get_static_method_id(klass, "parseFloat",'(Ljava/lang/String;)F')
        number = self.env.new_string_utf("-55.25")
        self.assertAlmostEqual(self.env.call_static_method(klass, method_id, number), -55.25)
        
    def test_04_07_call_static_double(self):
        klass = self.env.find_class("java/lang/Double")
        method_id = self.env.get_static_method_id(klass, "parseDouble",'(Ljava/lang/String;)D')
        number = self.env.new_string_utf("55.6")
        self.assertAlmostEqual(self.env.call_static_method(klass, method_id, number), 55.6)
        
    def test_04_08_call_static_object(self):
        klass = self.env.find_class("java/lang/String")
        method_id = self.env.get_static_method_id(klass, "valueOf",'(Z)Ljava/lang/String;')
        result = self.env.call_static_method(klass, method_id, True)
        self.assertEqual(self.env.get_string_utf(result), "true")
        
    def test_04_09_call_static_char(self):
        klass = self.env.find_class("java/lang/Character")
        method_id = self.env.get_static_method_id(klass, "toLowerCase", "(C)C")
        result = self.env.call_static_method(klass, method_id, "X")
        self.assertEqual(result, "x")
        
    def test_04_10_call_static_array(self):
        jstring = self.env.new_string_utf("Hello, world")
        klass = self.env.get_object_class(jstring)
        method_id = self.env.get_method_id(klass, "toCharArray","()[C")
        chars = self.env.call_method(jstring, method_id)
        method_id = self.env.get_static_method_id(klass, "copyValueOf","([C)Ljava/lang/String;")
        result = self.env.call_static_method(klass, method_id, chars)
        self.assertEqual(self.env.get_string_utf(result), "Hello, world")
    
    def test_05_01_get_static_object_field(self):
        klass = self.env.find_class("java/lang/Boolean")
        field_id = self.env.get_static_field_id(klass, "FALSE","Ljava/lang/Boolean;")
        result = self.env.get_static_object_field(klass, field_id)
        method_id = self.env.get_method_id(klass, "booleanValue","()Z")
        self.assertFalse(self.env.call_method(result, method_id))
        
    def test_05_02_get_static_boolean_field(self):
        pass # can't find any examples
    
    def test_05_03_get_static_byte_field(self):
        klass = self.env.find_class("java/io/ObjectStreamConstants")
        field_id = self.env.get_static_field_id(klass, "SC_EXTERNALIZABLE","B")
        result = self.env.get_static_byte_field(klass, field_id)
        self.assertEqual(result, 4)
        
    def test_05_04_get_static_short_field(self):
        klass = self.env.find_class("java/io/ObjectStreamConstants")
        field_id = self.env.get_static_field_id(klass, "STREAM_MAGIC","S")
        result = self.env.get_static_short_field(klass, field_id)
        self.assertEqual(result, -21267) # 0xaced see http://java.sun.com/javase/6/docs/platform/serialization/spec/protocol.html
    
    def test_05_05_get_static_int_field(self):
        klass = self.env.find_class("java/io/ObjectStreamConstants")
        field_id = self.env.get_static_field_id(klass, "PROTOCOL_VERSION_1","I")
        result = self.env.get_static_int_field(klass, field_id)
        self.assertEqual(result, 1)
        
    def test_05_06_get_static_long_field(self):
        klass = self.env.find_class("java/security/Key")
        field_id = self.env.get_static_field_id(klass, "serialVersionUID", "J")
        result = self.env.get_static_long_field(klass, field_id)
        self.assertEqual(result, 6603384152749567654)
    
    def test_05_07_get_static_float_field(self):
        klass = self.env.find_class("java/lang/Float")
        field_id = self.env.get_static_field_id(klass, "MAX_VALUE","F")
        result = self.env.get_static_float_field(klass, field_id)
        self.assertAlmostEqual(result, 3.4028234663852886 * 10.0**38)
        
    def test_05_08_get_static_double_field(self):
        klass = self.env.find_class("java/lang/Math")
        field_id = self.env.get_static_field_id(klass, "PI","D")
        result = self.env.get_static_double_field(klass, field_id)
        self.assertAlmostEqual(result, 3.141592653589793)
        
    def test_06_01_class_as_object(self):
        klass_map = self.env.find_class("java/util/Map")
        klass_map_obj = klass_map.as_class_object()
        klass_klass = self.env.find_class("java/lang/Class")
        method_getname = self.env.get_method_id(
            klass_klass, "getName", "()Ljava/lang/String;")
        jstring = self.env.call_method(klass_map_obj, method_getname)
        self.assertEqual(self.env.get_string_utf(jstring), "java.util.Map")



"""__init__.py - the javabridge package

python-javabridge is licensed under the BSD license.  See the
accompanying file LICENSE for details.

Copyright (c) 2003-2009 Massachusetts Institute of Technology
Copyright (c) 2009-2013 Broad Institute
All rights reserved.

"""

import os.path

try:
    from _version import __version__
except ImportError:
    # We're running in a tree that doesn't have a _version.py, so we don't know what our version is.
    __version__ = "0.0.0"

# We must dynamically find libjvm.so since its unlikely to be in the same place
# as it was on the distribution on which javabridge was built.
import sys
if sys.platform.startswith('linux'):
    from .locate import find_jre_bin_jdk_so
    _, jdk_so = find_jre_bin_jdk_so()
    if jdk_so:
        import ctypes
        ctypes.cdll.LoadLibrary(jdk_so)

_jars_dir = os.path.join(os.path.dirname(__file__), 'jars')

#: List of absolute paths to JAR files that are required for the
#: Javabridge to work.
JARS = [os.path.realpath(os.path.join(_jars_dir, name + '.jar'))
        for name in ['rhino-1.7R4', 'runnablequeue', 'cpython']]


from .jutil import start_vm, kill_vm, vm, activate_awt, deactivate_awt

from .jutil import attach, detach, get_env


# JavaScript
from .jutil import run_script, unwrap_javascript


# Operations on Java objects
from .jutil import call, get_static_field, static_call, \
    is_instance_of, make_instance, set_static_field, to_string, \
    get_field, set_field, make_static_call

# Make Python object that wraps a Java object
from .jutil import make_method, make_new, make_call, box
from .wrappers import JWrapper, JClassWrapper, JProxy

from .jutil import get_nice_arg

# Useful collection wrappers
from .jutil import get_dictionary_wrapper, jdictionary_to_string_dictionary, \
    jenumeration_to_string_list, get_enumeration_wrapper, iterate_collection, \
    iterate_java, make_list, get_collection_wrapper, make_future_task, \
    make_map, get_map_wrapper

# Reflection. (These use make_method or make_new internally.)
from .jutil import get_class_wrapper, get_field_wrapper, class_for_name, \
    get_constructor_wrapper, get_method_wrapper

# Ensure that callables, runnables and futures that use AWT run in the
# AWT main thread, which is not accessible from Python.
from .jutil import execute_callable_in_main_thread, \
    execute_runnable_in_main_thread, execute_future_in_main_thread, \
    get_future_wrapper

# Exceptions
from .jutil import JavaError, JavaException, JVMNotFoundError

from ._javabridge import mac_enter_run_loop, mac_stop_run_loop, mac_run_loop_init
    
# References
from .jutil import create_jref, redeem_jref, create_and_lock_jref,\
     lock_jref, unlock_jref

# Don't expose: AtExit, get_nice_args,
# make_run_dictionary, run_in_main_thread, split_sig, unwrap_javascript,
# print_all_stack_traces


# Low-level API
from ._javabridge import JB_Env, JB_Object, JB_Class
# JNI helpers.
from ._javabridge import jni_enter, jni_exit, jvm_enter



"""noseplugin.py - start and stop JVM when running unit tests

python-javabridge is licensed under the BSD license.  See the
accompanying file LICENSE for details.

Copyright (c) 2003-2009 Massachusetts Institute of Technology
Copyright (c) 2009-2013 Broad Institute
All rights reserved.

"""

import logging
from nose.plugins import Plugin
import os
import numpy as np
np.seterr(all='ignore')
import sys


log = logging.getLogger(__name__)


class JavabridgePlugin(Plugin):
    '''Javabridge nose test plugin
    
    This plugin starts the JVM before running tests and kills it when
    the tests are done. The plugin is necessary because the JVM cannot
    be restarted once it is killed, so unittest's setUp() and
    tearDown() methods cannot be used to start and stop the JVM.
    '''
    enabled = False
    name = "javabridge"
    score = 100
    extra_jvm_args = []

    def begin(self):
        import javabridge

        javabridge.start_vm(self.extra_jvm_args,
                            class_path=self.class_path.split(os.pathsep),
                            run_headless=self.headless,
                            max_heap_size=self.max_heap_size)
        if not self.headless:
            javabridge.activate_awt()

    def options(self, parser, env=os.environ):
        super(JavabridgePlugin, self).options(parser, env=env)
        parser.add_option("--classpath", action="store",
                          default=env.get('NOSE_CLASSPATH'),
                          metavar="PATH",
                          dest="classpath",
                          help="Additional class path for JVM [NOSE_CLASSPATH]")
        parser.add_option("--no-headless", action="store_true",
                          default=bool(env.get('NOSE_NO_HEADLESS')),
                          dest="no_headless",
                          help="Set Java environment variable, java.awt.headless to false to allow AWT calls [NOSE_NO_HEADLESS]")
        parser.add_option("--max-heap-size", action="store",
                          default=env.get('NOSE_MAX_HEAP_SIZE'),
                          dest="max_heap_size",
                          help="Set the maximum heap size argument to the JVM as in the -Xmx command-line argument [NOSE_MAX_HEAP_SIZE]")

    def configure(self, options, conf):
        import javabridge
        super(JavabridgePlugin, self).configure(options, conf)
        self.class_path = os.pathsep.join(javabridge.JARS)
        if options.classpath:
            self.class_path = os.pathsep.join([options.classpath, self.class_path])
        self.headless = not options.no_headless
        self.max_heap_size = options.max_heap_size

    def prepareTestRunner(self, testRunner):
        '''Need to make the test runner call finalize if in Wing
        
        Wing IDE's XML test runner fails to call finalize, so we
        wrap it and add that function here
        '''
        if (getattr(testRunner, "__module__","unknown") == 
            "wingtest_common"):
            outer_self = self
            class TestRunnerProxy(object):
                def run(self, test):
                    result = testRunner.run(test)
                    outer_self.finalize(testRunner.result)
                    return result
                
                @property
                def result(self):
                    return testRunner.result
            return TestRunnerProxy()
            
    def finalize(self, result):
        import javabridge
        javabridge.kill_vm()



'''test_wrappers.py test JWrapper and JClassWrapper

python-javabridge is licensed under the BSD license.  See the
accompanying file LICENSE for details.

Copyright (c) 2003-2009 Massachusetts Institute of Technology
Copyright (c) 2009-2013 Broad Institute
All rights reserved.

'''
import unittest
import javabridge as J

class TestJWrapper(unittest.TestCase):
    def test_01_01_init(self):
        jobj = J.get_env().new_string(u"Hello, world.")
        obj = J.JWrapper(jobj)
        self.assertEquals(jobj, obj.o)

    def test_01_02_call_noargs(self):
        jobj = J.get_env().new_string(u"Hello, world.")
        obj = J.JWrapper(jobj)
        self.assertEquals(obj.toLowerCase(), "hello, world.")

    def test_01_03_call_args(self):
        jobj = J.get_env().new_string(u"Hello, world.")
        obj = J.JWrapper(jobj)
        result = obj.replace("Hello,", "Goodbye cruel")
        self.assertEquals(result, "Goodbye cruel world.")

    def test_01_04_call_varargs(self):
        sclass = J.JWrapper(J.class_for_name("java.lang.String"));
        for constructor in J.get_env().get_object_array_elements(
            sclass.getConstructors().o):
            wconstructor = J.JWrapper(constructor)
            parameter_types = J.get_env().get_object_array_elements(
                wconstructor.getParameterTypes().o)
            c1 = sclass.getConstructor(*parameter_types)
            self.assertTrue(c1.equals(constructor))

    def test_02_01_get_field(self):
        obj = J.JClassWrapper("org.cellprofiler.javabridge.test.RealRect")(
            1.5, 2.5, 3.5, 4.5)
        self.assertEquals(obj.x, 1.5)

    def test_02_02_set_field(self):
        obj = J.JClassWrapper("org.cellprofiler.javabridge.test.RealRect")(
            1.5, 2.5, 3.5, 4.5)
        obj.x = 2.5
        self.assertEquals(obj.x, 2.5)

class TestJClassWrapper_Unboxing(unittest.TestCase):
    def setUp(self):
        self.i = J.JClassWrapper('java.lang.Integer')(3)

    def test_01_01_int(self):
        self.assertEquals(int(self.i), 3)
    
    def test_01_02_float(self):
        self.assertEquals(float(self.i),3.0)
    
    def test_01_03_str(self):
        self.assertEquals(str(self.i), '3')

class TestJClassWrapper_Collection(unittest.TestCase):
    def setUp(self):
        self.a = J.JClassWrapper('java.util.ArrayList')()
        self.assertEquals(len(self.a), 0)
        self.ints = [0,1,2,4,8,16]
        self.assertEquals(len(self.ints), 6)
        for i in self.ints:
            self.a.add(i)

    def test_01_01_get_len(self):
        self.assertEquals(len(self.a), len(self.ints))

    def test_01_02_iterate(self):
        for x,y in zip(self.a, self.ints):
            self.assertEquals(x.intValue(), y)

    def test_01_03_get_index(self):
        for i in range(len(self.a)):
            self.assertEquals(self.a[i].intValue(), self.ints[i])

    def test_01_04_set_index(self):
        for i in range(len(self.a)):
            self.a[i] = 10
        for i in range(len(self.a)):
            self.assertEquals(self.a[i].intValue(), 10)

class TestJClassWrapper(unittest.TestCase):
    def test_01_01_init(self):
        c = J.JClassWrapper("java.lang.Integer")

    def test_01_02_field(self):
        c = J.JClassWrapper("java.lang.Short")
        field = c.MAX_VALUE
        self.assertEquals(field, (1 << 15)-1)

    def test_02_03_static_call(self):
        c = J.JClassWrapper("java.lang.Integer")
        self.assertEquals(c.toString(123), "123")

    def test_02_04_static_call_varargs(self):
        #
        # Test calling a static function with a variable number of
        # arguments.
        #
        c = J.JClassWrapper("java.lang.String")
        self.assertEquals(c.format("Hello, %s.", "world"),
                                   "Hello, world.")
        self.assertEquals(c.format("Goodbye %s %s.", "cruel", "world"),
                          "Goodbye cruel world.")

    def test_02_05_constructor_varargs(self):
        # Regression test of issue #41
        #
        args = ("foo", "bar")
        f = J.JClassWrapper(
            "javax.swing.filechooser.FileNameExtensionFilter")("baz", *args)
        exts = J.get_env().get_object_array_elements(f.getExtensions().o)
        self.assertEqual(args[0], J.to_string(exts[0]))
        self.assertEqual(args[1], J.to_string(exts[1]))

class TestJProxy(unittest.TestCase):
    def test_01_01_init(self):
        def whatever():
            pass
        J.JProxy('java.lang.Runnable', dict(run=whatever))

    def test_01_02_runnable(self):
        magic = []
        def whatever(magic=magic):
            magic.append("bus")
        runnable = J.JProxy('java.lang.Runnable',
                            dict(run=whatever))
        J.JWrapper(runnable.o).run()
        self.assertEqual(magic[0], "bus")

    def test_01_03_runnable_class(self):
        class MyProxy(J.JProxy):
            def __init__(self):
                J.JProxy.__init__(self, 'java.lang.Runnable')
                self.esteem = 0

            def run(self):
                self.esteem = "through the roof"

        proxy = MyProxy()
        J.JWrapper(proxy.o).run()
        self.assertEqual(proxy.esteem, "through the roof")

    def test_01_04_args(self):
        my_observable = J.make_instance("java/util/Observable", "()V")
        def update(observable, obj):
            self.assertTrue(J.JWrapper(observable).equals(my_observable))
            self.assertTrue(J.JWrapper(obj).equals("bar"))
        proxy = J.JProxy('java.util.Observer', dict(update=update))
        J.JWrapper(proxy.o).update(my_observable, "bar")

    def test_01_05_return_value(self):
        def call():
            return "foo"
        proxy = J.JProxy('java.util.concurrent.Callable',
                         dict(call = call))
        self.assertEquals(J.JWrapper(proxy.o).call(), "foo")

if __name__=="__main__":
    import javabridge
    javabridge.start_vm()
    try:
        unittest.main()
    finally:
        javabridge.kill_vm()



"""setup.py - build python-javabridge

python-javabridge is licensed under the BSD license.  See the
accompanying file LICENSE for details.

Copyright (c) 2003-2009 Massachusetts Institute of Technology
Copyright (c) 2009-2013 Broad Institute
All rights reserved.

"""
from __future__ import print_function

import errno
import glob
import os
import re
import sys
try:
    import sysconfig
except:
    import distutils.sysconfig as sysconfig
import subprocess
import traceback
import distutils.log
from distutils.errors import DistutilsSetupError, DistutilsExecError, LinkError
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as _build_ext
from distutils.command.build_clib import build_clib
from distutils.ccompiler import CCompiler


# Hack to avoid importing the javabridge package
sys.path.append(os.path.join(os.path.dirname(__file__), 'javabridge'))
from locate import *

def in_cwd(basename):
    return os.path.join(os.path.dirname(__file__), basename)

def build_cython():
    """Compile the pyx files if we have them.

    The git repository has the .pyx files but not the .c files, and
    the source distributions that are uploaded to PyPI have the .c
    files and not the .pyx files. (The reason for the latter is that
    some versions of pip discovers the .pyx files and implicitly adds
    a dependency on Cython.) Therefore, if we have the .pyx files,
    compile them.

    """
    stems = ['_javabridge', '_javabridge_mac', '_javabridge_nomac']
    pyx_filenames = [in_cwd(s + '.pyx') for s in stems]
    c_filenames = [in_cwd(s + '.c') for s in stems]
    nc_pyx_filenames = [
        pyx for pyx, c in zip(pyx_filenames, c_filenames)
        if os.path.exists(pyx) and needs_compilation(c, pyx)]
    if len(nc_pyx_filenames) > 0:
        cython_cmd = [sys.executable, '-m', 'cython', "-3"]
        cmd = cython_cmd + nc_pyx_filenames
        env = dict(os.environ)
        env['PYTHONPATH'] = os.pathsep.join(sys.path)
        try:
            subprocess.check_call(cmd, env=env)
        except FileNotFoundError:
            raise RuntimeError("Failed to find Cython: {}".format(cython_cmd))

def get_jvm_include_dirs():
    '''Return a sequence of paths to include directories for JVM defs'''
    jdk_home = find_jdk()
    java_home = find_javahome()
    include_dirs = []
    if is_win:
        if jdk_home is not None:
            jdk_include = os.path.join(jdk_home, "include")
            jdk_include_plat = os.path.join(jdk_include, sys.platform)
            include_dirs += [jdk_include, jdk_include_plat]
    elif is_mac:
        where_jni_h_is_post_6 = os.path.join(java_home, 'include')
        if os.path.isfile(os.path.join(where_jni_h_is_post_6, "jni.h")):

            include_dirs += [where_jni_h_is_post_6,
                             os.path.join(java_home, 'include', 'darwin')]
        else:
            include_dirs += ["/System/Library/Frameworks/JavaVM.Framework/Headers"]
    elif is_linux:
        include_dirs += [os.path.join(jdk_home,'include'),
                         os.path.join(jdk_home,'include','linux'),
                         os.path.join(jdk_home,'default-java','include'),
                         os.path.join(jdk_home,'default-java','include','linux')
                         ]

    return include_dirs

def ext_modules():
    extensions = []
    extra_link_args = None
    java_home = find_javahome()
    if java_home is None:
        raise Exception("JVM not found")
    jdk_home = find_jdk()
    include_dirs = get_jvm_include_dirs()
    libraries = None
    library_dirs = None
    javabridge_sources = ['_javabridge.c']
    _, jvm_so = find_jre_bin_jdk_so()
    if is_mac:
        javabridge_sources += ['_javabridge_mac.c']
        extra_link_args = ['-framework', 'CoreFoundation']
    else:
        javabridge_sources += ['_javabridge_nomac.c']
    if is_win:
        jdk_lib = os.path.join(jdk_home, "lib")
        if is_mingw:
            #
            # Build libjvm from jvm.dll on Windows.
            # This assumes that we're using mingw32 for build
            #
	    # generate the jvm.def file matching to the jvm.dll
            cmd = ["gendef", os.path.join(jdk_home,"jre\\bin\\server\\jvm.dll")]
            p = subprocess.Popen(cmd)
            p.communicate()
            cmd = ["dlltool", "--dllname",
                   jvm_so,
                   "--output-lib","libjvm.a",
                   "--input-def","jvm.def",
                   "--kill-at"]
            p = subprocess.Popen(cmd)
            p.communicate()
            library_dirs = [os.path.abspath("."), jdk_lib]
        else:
            #
            # Use the MSVC lib in the JDK
            #
            extra_link_args = ['/MANIFEST']
            library_dirs = [jdk_lib]
            javabridge_sources.append("strtoull.c")

        libraries = ["jvm"]
    elif is_mac:
        javabridge_sources += [ "mac_javabridge_utils.c" ]
    elif is_linux:
        library_dirs = [os.path.dirname(jvm_so)]
        libraries = ["jvm"]
    extension_kwargs = dict(
        name="javabridge._javabridge",
        sources=javabridge_sources,
        libraries=libraries,
        library_dirs=library_dirs,
        include_dirs=include_dirs,
        extra_link_args=extra_link_args)
    if not is_win:
        extension_kwargs["runtime_library_dirs"] =library_dirs

    extensions += [Extension(**extension_kwargs)]
    return extensions

SO = ".dll" if sys.platform == 'win32' \
    else ".jnilib" if sys.platform == 'darwin'\
    else ".so"

def needs_compilation(target, *sources):
    try:
        target_date = os.path.getmtime(target)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
        return True
    for source in sources:
        source_date = os.path.getmtime(source)
        if source_date > target_date:
            return True
    return False

def package_path(relpath):
    return os.path.normpath(os.path.join(os.path.dirname(__file__), relpath))

class build_ext(_build_ext):
    java2cpython_sources = ["java/org_cellprofiler_javabridge_CPython.c"]

    def initialize_options(self):
        from numpy import get_include
        _build_ext.initialize_options(self)
        if self.include_dirs is None:
            self.include_dirs = get_include()
        else:
            self.include_dirs += get_include()

    def run(self, *args, **kwargs):
        self.build_java()
        result = build_cython()
        if self.inplace:
            dirty = False
            for source in self.get_source_files():
                source_mtime = os.stat(source).st_mtime
                for output in self.get_outputs():
                    if not os.path.isfile(output) or \
                       os.stat(output).st_mtime < source_mtime:
                        dirty = True
                        break
            output_dir = os.path.splitext(
                self.get_ext_fullpath("javabridge.jars"))[0]
            java2cpython_lib = os.path.join(
                output_dir, self.get_java2cpython_libdest()[1])
            if (not os.path.exists(java2cpython_lib)) or \
               any([os.stat(src).st_mtime > os.stat(java2cpython_lib).st_mtime
                    for src in self.java2cpython_sources]):
                dirty = True
        else:
            dirty = True
        if dirty:
            result = _build_ext.run(self, *args, **kwargs)
            self.build_java2cpython()
        return result

    def build_jar_from_sources(self, jar, sources):
        if sys.platform == 'win32':
            sources = [source.replace("/", os.path.sep) for source in sources]
        jar_filename = jar.rsplit(".", 1)[1] + ".jar"
        jar_dir = os.path.dirname(self.get_ext_fullpath(jar))
        jar = os.path.join(jar_dir, jar_filename)
        jar_command = [find_jar_cmd(), 'cf', package_path(jar)]

        javac_loc = find_javac_cmd()
        dirty_jar = False
        javac_command = [javac_loc]
        for source in sources:
            javac_command.append(package_path(source))
            if needs_compilation(jar, source):
                dirty_jar = True

        self.spawn(javac_command)
        if dirty_jar:
            if not os.path.exists(os.path.dirname(jar)):
                os.mkdir(os.path.dirname(jar))
            for source in sources:
                for klass in glob.glob(source[:source.rindex('.')] + '*.class'):
                    java_klass_path = klass[klass.index(os.path.sep) + 1:].replace(os.path.sep, "/")
                    jar_command.extend(['-C', package_path('java'), java_klass_path])
            self.spawn(jar_command)

    def build_java2cpython(self):
        sources = self.java2cpython_sources
        distutils.log.info("building java2cpython library")


        # First, compile the source code to object files in the library
        # directory.  (This should probably change to putting object
        # files in a temporary build directory.)
        include_dirs = \
            [sysconfig.get_config_var("INCLUDEPY"), "java"] +\
            get_jvm_include_dirs()
        python_lib_dir, lib_name = self.get_java2cpython_libdest()
        library_dirs = [python_lib_dir]
        output_dir = os.path.join(os.path.dirname(
            self.get_ext_fullpath("javabridge.jars")), "jars")
        export_symbols = ['Java_org_cellprofiler_javabridge_CPython_exec']
        objects = self.compiler.compile(sources,
                                        output_dir=self.build_temp,
                                        include_dirs=include_dirs,
                                        debug=self.debug)
        ver = sys.version_info
        needs_manifest = sys.platform == 'win32' and ver.major == 2 and not is_mingw
        extra_postargs = ["/MANIFEST"] if needs_manifest else None
        libraries = ["python{}{}".format(ver.major, ver.minor)] if is_mingw else None
        self.compiler.link(
            CCompiler.SHARED_OBJECT,
            objects, lib_name,
            output_dir=output_dir,
            debug=self.debug,
            library_dirs=library_dirs,
	    libraries=libraries,
            export_symbols=export_symbols,
            extra_postargs=extra_postargs)
        if needs_manifest:
            temp_dir = os.path.dirname(objects[0])
            manifest_name = lib_name +".manifest"
            lib_path = os.path.join(output_dir, lib_name)
            manifest_file = os.path.join(temp_dir, manifest_name)
            lib_path = os.path.abspath(lib_path)
            manifest_file = os.path.abspath(manifest_file)
            out_arg = '-outputresource:%s;2' % lib_path
            try:
                self.compiler.spawn([
                    'mt.exe', '-nologo', '-manifest', manifest_file,
                    out_arg])
            except DistutilsExecError as msg:
                raise LinkError(msg)

    def get_java2cpython_libdest(self):
        if is_win:
            python_lib_dir = os.path.join(
                sysconfig.get_config_var('platbase'),
                'LIBS')
            lib_name = "java2cpython" + SO
        else:
            python_lib_dir = sysconfig.get_config_var('LIBDIR')
            lib_name = "libjava2cpython" + SO
        return python_lib_dir, lib_name


    def build_jar_from_single_source(self, jar, source):
        self.build_jar_from_sources(jar, [source])

    def build_runnablequeue(self):
        jar = 'javabridge.jars.runnablequeue'
        source = 'java/org/cellprofiler/runnablequeue/RunnableQueue.java'
        self.build_jar_from_single_source(jar, source)

    def build_cpython(self):
        jar = 'javabridge.jars.cpython'
        sources = [
            'java/org/cellprofiler/javabridge/CPython.java',
            'java/org/cellprofiler/javabridge/CPythonInvocationHandler.java']
        self.build_jar_from_sources(jar, sources)

    def build_test(self):
        jar = 'javabridge.jars.test'
        source = 'java/org/cellprofiler/javabridge/test/RealRect.java'
        self.build_jar_from_single_source(jar, source)

    def build_java(self):
        self.build_runnablequeue()
        self.build_test()
        self.build_cpython()


def pep440_compliant(ver):
    if ver is None:
        return ver
    m = re.match(r"^(?P<version>(\d[\d\.]*))$", ver)
    if m:
        return ver
    m = re.match(r"^(?P<version>(\d[\d\.]*))-(?P<count>\d+)-(?P<sha>.*)$", ver)
    if m:
        res = m.group('version') + '.post' + m.group('count') + '+' + m.group('sha')
        return res
    return ver


def get_version():
    """Get version from git or file system.

    If this is a git repository, try to get the version number by
    running ``git describe``, then store it in
    javabridge/_version.py. Otherwise, try to load the version number
    from that file. If both methods fail, quietly return None.

    """
    git_version = None
    if os.path.exists(os.path.join(os.path.dirname(__file__), '.git')):
        import subprocess
        try:
            git_version = subprocess.Popen(['git', 'describe'],
                                           stdout=subprocess.PIPE).communicate()[0].strip().decode('utf-8')
        except:
            pass

    version_file = os.path.join(os.path.dirname(__file__), 'javabridge',
                                '_version.py')
    if os.path.exists(version_file):
        with open(version_file) as f:
            cached_version_line = f.read().strip()
        try:
            # From http://stackoverflow.com/a/3619714/17498
            cached_version = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                                       cached_version_line, re.M).group(1)
        except:
            raise RuntimeError("Unable to find version in %s" % version_file)
    else:
        cached_version = None

    if git_version and git_version != cached_version:
        with open(version_file, 'w') as f:
            print('__version__ = "%s"' % git_version, file=f)

    return pep440_compliant(git_version or cached_version)


if __name__ == '__main__':
    if '/' in __file__:
        os.chdir(os.path.dirname(__file__))

    setup(name="javabridge",
          version=get_version(),
          description="Python wrapper for the Java Native Interface",
          long_description='''The python-javabridge package makes it easy to start a Java virtual
machine (JVM) from Python and interact with it. Python code can
interact with the JVM using a low-level API or a more convenient
high-level API. Python-javabridge was developed for and is used by the
cell image analysis software CellProfiler (cellprofiler.org).''',
          url="http://github.com/CellProfiler/python-javabridge/",
          packages=['javabridge', 'javabridge.tests'],
          classifiers=['Development Status :: 5 - Production/Stable',
                       'License :: OSI Approved :: BSD License',
                       'Programming Language :: Java',
                       'Programming Language :: Python :: 2',
                       'Programming Language :: Python :: 3'
                       ],
          license='BSD License',
          setup_requires=['cython', 'numpy'],
          install_requires=['numpy'],
          tests_require="nose",
          entry_points={'nose.plugins.0.10': [
                'javabridge = javabridge.noseplugin:JavabridgePlugin'
                ]},
          test_suite="nose.collector",
          ext_modules=ext_modules(),
          package_data={"javabridge": [
              'jars/*.jar', 'jars/*%s' % SO, 'VERSION']},
          cmdclass={'build_ext': build_ext})
